import os
from abc import ABC, abstractmethod
from pydantic import BaseModel as PydanticModel, ConfigDict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from langchain_core.messages import AIMessage, BaseMessage
    from langchain_core.language_models.chat_models import BaseChatModel


class EnvironmentAuth:
    env_settings: dict[str, str]

    def __init__(self, **env_settings: dict[str, str]) -> None:
        for key, value in env_settings.items():
            if not (isinstance(key, str) and isinstance(value, str)):
                raise ValueError("EnvironmentAuth variables names and values must be strings.")
        self.env_settings = env_settings

    def apply(self):
        os.environ.update(self.env_settings)

def set_env_auth(**env_settings: dict[str, str]) -> None:
    for key, value in env_settings.items():
        if not (isinstance(key, str) and isinstance(value, str)):
            raise ValueError("EnvironmentAuth variables names and values must be strings.")
    os.environ.update(env_settings)


class ModelConfig(PydanticModel, extra='allow'):
    model_name: str
    api_key: str

    model_config = ConfigDict(extra="allow", protected_namespaces=())


class BaseArchytasModel(ABC):

    MODEL_PROMPT_INSTRUCTIONS: str = ""

    model: "BaseChatModel"
    config: ModelConfig

    def __init__(self, config: ModelConfig, **kwargs) -> None:
        self.config = config
        self.auth(**kwargs)
        self.model = self.initialize_model(**kwargs)

    def auth(self, **kwargs) -> None:
        pass

    @property
    def additional_prompt_info(self) -> str | None:
        return None

    @abstractmethod
    def initialize_model(self, **kwargs):
        ...

    def invoke(self, input, *, config=None, stop=None, **kwargs):
        return self.model.invoke(
            self._preprocess_messages(input),
            config,
            stop=stop,
            **kwargs
        )

    async def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        try:
            messages = self._preprocess_messages(input)
            return await self.model.ainvoke(
                messages,
                config,
                stop=stop,
                **kwargs
            )
        except Exception as error:
            print(error)
            return self.handle_invoke_error(error)

    def _preprocess_messages(self, messages: "list[BaseMessage]"):
        return messages

    def _rectify_result(self, response_message: "AIMessage"):
        return response_message

    def process_result(self, response_message: "AIMessage"):
        content = response_message.content
        if isinstance(content, list):
            return "\n".join(item['text'] for item in content if item.get('type', None) == "text")
        return content

    def handle_invoke_error(self, error: BaseException):
        raise error



import json
from typing import TYPE_CHECKING
from langchain_google_genai.chat_models import ChatGoogleGenerativeAI, ChatGoogleGenerativeAIError

from .base import BaseArchytasModel
from ..message_schemas import ToolUseRequest
from ..exceptions import AuthenticationError, ExecutionError

if TYPE_CHECKING:
    from ..agent import SystemMessage, AutoContextMessage, AIMessage, ToolMessage, FunctionMessage


class GeminiModel(BaseArchytasModel):
    api_key: str

    def auth(self, **kwargs) -> None:
        auth_token = None
        if 'api_key' in kwargs:
            auth_token = kwargs['api_key']
        elif 'api_key' in self.config:
            auth_token = self.config['api_key']
        if auth_token:
            self.api_key = auth_token
        else:
            raise AuthenticationError("Gemini API key not provided.")

    def initialize_model(self, **kwargs):
        from langchain_google_genai._function_utils import _dict_to_gapic_schema
        schema = ToolUseRequest.model_json_schema()
        for property in schema["properties"].values():
            if "type" not in property:
                property["type"] = "string"
                property["description"] += " (As a JSON object encoded as a string)"
        schema = _dict_to_gapic_schema(schema)
        return ChatGoogleGenerativeAI(model=self.config.get("model_name", "gpt-4o"), api_key=self.api_key).bind(
                tool_config={
                    "function_calling_config": {
                        "mode": "NONE",
                    }
                },
                generation_config={
                    "response_mime_type": "application/json",
                    "response_schema": schema,
                },
            )

    async def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        # Gemini doesn't accept a temperature keyword on invoke
        kwargs.pop("temperature")
        return await super().ainvoke(input, config=config, stop=stop, **kwargs)

    def _preprocess_messages(self, messages):
        from ..agent import SystemMessage, AutoContextMessage
        output = []
        system_messages = []
        for message in messages:
            if isinstance(message, (SystemMessage, AutoContextMessage)):
                system_messages.append(message.content)
            else:
                output.append(message)
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        return output

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, ChatGoogleGenerativeAIError):
            if any(('400 API key not valid' in arg for arg in error.args)):
                raise AuthenticationError("API key invalid.") from error
        raise ExecutionError(*error.args) from error

    def process_result(self, response_message: "AIMessage"):
        content = response_message.content
        try:
            if isinstance(content, str):
                content = json.loads(content)
            content["tool_input"] = json.loads(content["tool_input"])
        except json.JSONDecodeError:
            pass
        return content



import asyncio
import json
import sys
import logging
import typing
import uuid
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage, ToolCall

from .agent import Agent, BaseMessage, SystemMessage
from .prompt import build_prompt, build_all_tool_names
from .tools import ask_user
from .tool_utils import make_tool_dict, sanitize_toolname
from .models.base import BaseArchytasModel
from .utils import extract_json


logger = logging.Logger("archytas")


class Undefined:
    ...


class FailedTaskError(Exception):
    def __init__(self, message: str, last_error: typing.Optional[BaseException] = None, tool_call_id: typing.Optional[str] = None, *extra_args: object) -> None:
        self.message = message
        self.last_error = last_error
        self.tool_call_id = tool_call_id
        super().__init__(*extra_args)

    def __str__(self) -> str:
        result = f"Error completing task: {self.message}"
        if self.last_error:
            result += f"\n\nThe last error is as follows:\n{str(self.last_error)}"
        return result


class LoopController:
    PROCEED = 0
    STOP_SUCCESS = 1
    STOP_FATAL = 2

    state: int

    def __init__(self) -> None:
        self.reset()

    def set_state(self, new_value):
        self.state = new_value

    def reset(self):
        self.state = 0


class AutoSummarizedToolMessage(SystemMessage):
    """A message that replaces its full tool output with a summary after the ReAct loop is complete."""

    summary_content: str
    summarized: bool

    def __init__(
        self,
        tool_content: str,
        summary_content: str,
    ):
        self.summarized = False
        self.summary_content = summary_content
        super().__init__(tool_content)

    async def update_content(self):
        self.update(content=self.summary_content)
        self.summarized = True


class ReActAgent(Agent):
    def __init__(
        self,
        *,
        model: BaseArchytasModel | None = None,
        api_key: str | None = None,
        tools: list = None,
        allow_ask_user: bool = True,
        max_errors: int | None = 3,
        max_react_steps: int | None = None,
        thought_handler: typing.Callable | None = Undefined,
        messages: typing.Optional[list[BaseMessage]] | None = None,
        **kwargs,
    ):
        """
        Create a ReAct agent

        Args:
            model (BaseArchytasModel): The model to use. Defaults to OpenAIModel(model_name="gpt-4o").
            api_key (str, optional): The LLM provider API key to use. Defaults to None. If None, the provider will use the default environment variable (e.g. OPENAI_API_KEY).
            tools (list): A list of tools to use. Defaults to None. If None, only the system tools (final_answer, fail_task) will be used.
            allow_ask_user (bool): Whether to include the ask_user tool, which allows the model to ask the user for clarification. Defaults to True.
            max_errors (int, optional): The maximum number of errors to allow during a task. Defaults to 3.
            max_react_steps (int, optional): The maximum number of steps to allow during a task. Defaults to infinity.
            thought_handler (function, optional): Hook to control logging/output of the thoughts made in the middle of a react loop. Set to None to disable, or leave default of Undefined to
                    print to terminal. Otherwise expects a callable function with the signature of `func(thought: str, tool_name: str, tool_input: str) -> None`.
            messages (list[BaseMessage], optional): A list of messages to start the agent with. Defaults to None.
        """
        # create a dictionary for looking up tools by name
        tools = tools or []
        if allow_ask_user:
            tools.append(ask_user)
        tools.append(self)
        self._raw_tools = tools
        self.tools = make_tool_dict(tools)
        self.current_query = None

        if thought_handler is Undefined:
            self.thought_handler = self.thought_callback
        else:
            self.thought_handler = thought_handler

        # check that the tools dict keys match the list of generated tool names
        names, keys = sorted(build_all_tool_names(tools)), sorted([*self.tools.keys()])
        assert (
            names == keys
        ), f"Internal Error: tools dict keys does not match list of generated tool names. {names} != {keys}"

        # create the prompt with the tools, and initialize the agent
        self.prompt = build_prompt(tools)
        super().__init__(model=model, prompt=self.prompt, api_key=api_key, messages=messages, **kwargs)

        # react settings
        self.max_errors = max_errors or float("inf")
        self.max_react_steps = max_react_steps or float("inf")

        # number of errors and steps during current task
        self.errors = 0
        self.steps = 0

        # keep track of the last tool used (for error messages)
        self.last_tool_name = ""

    def update_prompt(self):
        self.prompt = build_prompt(self._raw_tools)
        if self.model.MODEL_PROMPT_INSTRUCTIONS:
            self.prompt += "\n\n" + self.model.MODEL_PROMPT_INSTRUCTIONS
        self.system_message = SystemMessage(content=self.prompt)

    def disable(self, *tool_names):
        if len(tool_names) == 0:
            return
        for tool_name in tool_names:
            if tool_name in self.tools:
                setattr(self.tools[tool_name], '_disabled', True)
            elif "." not in tool_name:
                matches = [name for name in self.tools.keys() if name.endswith(f".{tool_name}")]
                if len(matches) > 1:
                    raise ValueError(f"Ambiguous name: Multiple tools called '{tool_name}'")
                elif len(matches) == 1:
                    subtool_name = matches[0]
                    method = self.tools[subtool_name]

                    setattr(method.__func__, '_disabled', True)

        self.update_prompt()


    def thought_callback(self, thought: str, tool_name: str, tool_input: str) -> None:
        if self.verbose:
            # TODO: better coloring
            self.print(
                f"thought: {thought}\ntool: {tool_name}\ntool_input: {tool_input}\n"
            )

    def react(self, query: str, react_context:dict=None) -> str:
        """
        Synchronous wrapper around the asynchronous react_async method.
        """
        return asyncio.run(self.react_async(query, react_context))

    async def react_async(self, query: str, react_context:dict=None) -> str:
        """
        Asynchronous react loop function.
        Continually calls tools until a satisfactory answer is reached.
        """

        # reset error and steps counter
        self.errors = 0
        self.steps = 0

        action = None
        reaction = None

        # Set the current query for use in tools, auto context, etc
        self.current_query = query
        action = await self.handle_message(HumanMessage(content=query))

        controller = LoopController()

        # ReAct loop
        while True:
            # Start processing last loops reaction as a new action
            if reaction is not None:
                action = reaction

            message = self.messages[-1]
            tool_id = uuid.uuid4().hex

            # Convert agent output to json
            if isinstance(action, str):
                try:
                    # First, assume it is a ```json ... ``` blocked string. If no blocks found (ValueError), try in case
                    # the action string is a valid JSON string.
                    try:
                        action = extract_json(action)
                    except ValueError as extract_err:
                        if "Unable to find json block" in extract_err.args:
                            action = json.loads(action)
                        else:
                            raise
                except json.JSONDecodeError as decode_err:
                    reaction = await self.error("Error parsing JSON", decode_err, tool_id=tool_id)
                    continue
                except Exception as e:
                    raise

            # verify that action has the correct keys
            try:
                thought, tool_name, tool_input, helpful_thought = self.extract_action(action)
                self.debug(
                    event_type="react_thought",
                    content={
                        "id": tool_id,
                        "thought": thought,
                        "tool_name": tool_name,
                        "tool_input": tool_input,
                        "helpful_thought": helpful_thought,
                    }
                )
                self.last_tool_name = tool_name  # keep track of the last tool used
            except AssertionError as e:
                reaction = await self.error("Error", e, tool_id=tool_id)
                continue

            if self.thought_handler:
                self.thought_handler(thought, tool_name, tool_input)

            # exit ReAct loop if agent says final_answer or fail_task
            if tool_name == "final_answer":
                await self.summarize_messages()
                self.debug(
                    event_type="react_final_answer",
                    content={
                        "final_answer": tool_input,
                    }
                )
                self.current_query = None
                # Store final answer as response in message,
                if not isinstance(tool_input, str):
                    tool_input = str(tool_input)
                self.messages.append(AIMessage(content=tool_input))
                return tool_input
            else:
                message.tool_calls.append(
                    ToolCall(
                        id=tool_id,
                        name=sanitize_toolname(tool_name),
                        args={"arg_string": tool_input},
                    )
                )
            if tool_name == "fail_task":
                self.current_query = None
                raise FailedTaskError(
                    f"Unable to complete the requested task. Giving up.\n Tool input: {tool_input}",
                    tool_call_id=tool_id
                )

            # run tool
            try:
                tool_fn = self.tools[tool_name]
            except KeyError:
                reaction = await self.error(f'Unknown tool "{tool_name}"\nAvailable tools: {", ".join(self.tools.keys())}', None, tool_id=tool_id)
                continue

            try:
                tool_context = {
                    "agent": self,
                    "tool_name": tool_name,
                    "raw_tool": tool_fn,
                    "loop_controller": controller,
                    "react_context": react_context,
                }
                tool_self_ref = getattr(tool_fn, "__self__", None)
                self.debug(
                    event_type="react_tool",
                    content={
                        "tool": tool_name,
                        "input": tool_input,
                    }
                )
                tool_output = await tool_fn.run(tool_input, tool_context=tool_context, self_ref=tool_self_ref)
                self.debug(
                    event_type="react_tool_output",
                    content={
                        "tool": tool_name,
                        "input": tool_input,
                        "output": tool_output,
                    }
                )
            except Exception as e:
                reaction = await self.error(f'error running tool "{tool_name}"', e, tool_id=tool_id)
                continue

            if controller.state != LoopController.PROCEED:
                self.debug(
                    event_type="react_controller_state",
                    content={
                        "state": controller.state,
                    }
                )
            # Check loop controller to see if we need to stop or error
            if controller.state == LoopController.STOP_SUCCESS:
                await self.summarize_messages()
                self.current_query = None
                return tool_output
            if controller.state == LoopController.STOP_FATAL:
                await self.summarize_messages()
                self.current_query = None
                raise FailedTaskError(
                    tool_output,
                    tool_call_id=tool_id
                )

            # have the agent observe the result, and get the next action
            if self.verbose:
                self.display_observation(tool_output)
            if getattr(tool_fn, "autosummarize", False):
                reaction = await self.handle_message(AutoSummarizedToolMessage(
                    tool_content=tool_output,
                    summary_content=f"Summary of action: Executed command '{tool_name}' with input '{tool_input}'",
                ))
            else:
                reaction = await self.handle_message(
                    ToolMessage(
                        content=tool_output,
                        tool_call_id=tool_id,
                    )
                )

    @staticmethod
    def extract_action(action: dict) -> tuple[str, str, str, bool]:
        """Verify that action has the correct keys. Otherwise, raise an error"""
        assert isinstance(
            action, dict
        ), f"Action must be a json dictionary, got {type(action)}"
        assert "thought" in action, "Action json is missing key 'thought'"
        assert "tool" in action, "Action json is missing key 'tool'"
        assert "tool_input" in action, "Action json is missing key 'tool_input'"

        thought = action["thought"]
        tool = action["tool"]
        tool_input = action["tool_input"]
        helpful_thought = action.get("helpful_thought", True)


        return thought, tool, tool_input, helpful_thought

    def execute(self, additional_messages: list[BaseMessage] = []) -> str:
        """
        Execute the model and return the output (see `Agent.execute()`).
        Keeps track of the number of execute calls, and raises an error if there are too many.
        """
        try:
            self.steps += 1
            if self.steps > self.max_react_steps:
                last_tool_id = None
                raise FailedTaskError(
                    f"Too many steps ({self.steps} > max_react_steps) during task.\nLast action should have been either final_answer or fail_task. Instead got: {self.last_tool_name}",
                    tool_call_id=last_tool_id,
                )
            result = super().execute(additional_messages)
        except FailedTaskError as err:
            # Ensure that the last message is a tool message, adding a new one if needed
            if not isinstance(self.messages[-1], ToolMessage):
                content = f"""\
Task failed:
{err.message}
"""
                if err.last_error:
                    content += f"""
During processing, the following error was raised:
{err.last_error}
"""
                error_message = ToolMessage(
                    content=content,
                    tool_call_id=err.tool_call_id,
                )
                self.messages.append(error_message)
            raise
        return result

    def error(self, mesg: str, err: BaseException, tool_id: str | None = None) -> str:
        """error handling. If too many errors, break the ReAct loop. Otherwise tell the agent, and continue"""
        error_message = ToolMessage(
            content=f"{mesg}: {err}",
            tool_call_id=tool_id,
        )

        self.errors += 1
        if self.errors >= self.max_errors:
            self.messages.append(error_message)
            raise FailedTaskError(
                f"Too many errors during task. Last error: {mesg}",
                last_error=err,
                tool_call_id=tool_id
            )
        if self.verbose:
            if self.rich_print:
                self.print(f"[red]error: {mesg}[/red]", file=sys.stderr)
            else:
                self.print(f"error: {mesg}", file=sys.stderr)

        # tell the agent about the error, and get its response (call parent .error method)
        return super().error(error_message)

    def display_observation(self, observation):
        """
        Display the observation. Can be overridden by subclasses to display the observation in different ways.
        """
        self.print(f"observation: {observation}\n")

    async def summarize_messages(self):
        """Summarizes and self-summarizing tool messages."""
        for message in self.messages:
            if isinstance(message, AutoSummarizedToolMessage) and not message.summarized:
                await message.update_content()



import re
from langchain_openai.chat_models import ChatOpenAI
from langchain_core.messages import FunctionMessage, AIMessage

from openai import AuthenticationError as OpenAIAuthenticationError, APIError, APIConnectionError, RateLimitError, OpenAIError
from .base import BaseArchytasModel, ModelConfig, set_env_auth
from ..exceptions import AuthenticationError, ExecutionError

DEFERRED_TOKEN_VALUE = "***deferred***"

class OpenAIModel(BaseArchytasModel):

    def auth(self, **kwargs) -> None:
        auth_token = None
        if 'api_key' in kwargs:
            auth_token = kwargs['api_key']
        elif 'api_key' in self.config:
            auth_token = self.config['api_key']
        if auth_token is None:
            auth_token = DEFERRED_TOKEN_VALUE
        set_env_auth(OPENAI_API_KEY=auth_token)

        if auth_token != DEFERRED_TOKEN_VALUE:
            self.config['api_key'] = auth_token
        # Reset the openai client with the new value, if needed.
        if getattr(self, "model", None):
            self.model.openai_api_key._secret_value = auth_token
            self.model.client = None
            self.model.async_client = None

            # This method reinitializes the clients
            self.model.validate_environment()

    def initialize_model(self, **kwargs):
        try:
            return ChatOpenAI(model=self.config.get("model_name", "gpt-4o"))
        except (APIConnectionError, OpenAIError) as err:
            if not self.config.get('api_key', None):
                raise AuthenticationError("OpenAI API Key not set")
            else:
                raise AuthenticationError("OpenAI Authentication Error") from err

    def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        if self.model is None or not getattr(self.model, 'openai_api_key', None):
            raise AuthenticationError("OpenAI API Key missing")
        return super().ainvoke(input, config=config, stop=stop, **kwargs)

    def _preprocess_messages(self, messages):
        output = []
        for message in messages:
            if isinstance(message, FunctionMessage):
                message.name = re.sub(r'[^a-zA-Z0-9_-]', '_', message.name)
            elif isinstance(message, AIMessage):
                for tool_call in message.tool_calls:
                    tool_call["name"] = re.sub(r'[^a-zA-Z0-9_-]', '_', tool_call["name"])
            output.append(message)
        return output

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, OpenAIAuthenticationError):
            raise AuthenticationError("OpenAI Authentication Error") from error
        elif isinstance(error, RateLimitError):
            raise ExecutionError(error.message) from error
        elif isinstance(error, (APIConnectionError, OpenAIError)) and not self.model.openai_api_key:
            raise AuthenticationError("OpenAI Authentication Error") from error
        else:

            raise error



from pydantic import BaseModel
from typing import Any, TYPE_CHECKING
if TYPE_CHECKING:
    from _typeshed import DataclassInstance  # only available to type checkers
from dataclasses import is_dataclass, asdict, _MISSING_TYPE

from .archytypes import (
    normalize_type, NormalizedType,
    Dataclass_t, PydanticModel_t,
    List_t, Dict_t, Tuple_t, Union_t,
    NotProvided,
    is_structured_type
)
from .constants import TAB



def verify_model_fields(cls: type[BaseModel], data: dict) -> dict:
    """
    Verify that the keys in the data dictionary are valid fields for the model class.
    Recursively verifies any nested models.
    """
    cls_fields = set(cls.__annotations__.keys())
    data_fields = set(data.keys())

    # Check if the provided data fields are all valid
    if not data_fields.issubset(cls_fields):
        raise ValueError(
            f"Invalid fields as input to {cls.__name__}: {data_fields - cls_fields}. Valid fields are: {cls_fields}")

    # Recursively verify nested models
    for key, value in data.items():
        if isinstance(value, dict) and issubclass(cls.__annotations__[key], BaseModel):
            # Recursively call verify_model_fields for nested models
            nested_cls = cls.__annotations__[key]
            verify_model_fields(nested_cls, value)

    return data


def construct_structured_type(arg_type: NormalizedType, data: dict) -> 'DataclassInstance|BaseModel|list':
    """
    construct an instance of a structured type from a dictionary (currently only support dataclass and pydantic models)
    recursively construct nested structured types, and validate the input data against the structured type

    Args:
        arg_type (type): The structured type to construct
        data (dict): The dictionary to construct the structured type from

    Returns:
        The constructed structured type instance 
    """

    if isinstance(arg_type, Dataclass_t):
        return construct_dataclass(arg_type, data)

    if isinstance(arg_type, PydanticModel_t):
        verify_model_fields(arg_type.cls, data)
        return arg_type.cls(**data)

    if isinstance(arg_type, List_t):
        assert isinstance(data, list), f"Expected list, got {type(data)} for input {data}"
        assert not isinstance(arg_type.element_type, NotProvided), f"INTERNAL ERROR: {arg_type} is not a structured type"
        return [construct_structured_type(arg_type.element_type, item) for item in data]

    if isinstance(arg_type, Union_t):
        return construct_union(arg_type, data)

    # Future support for other structured types can be added here
    raise ValueError(f"Unsupported structured type {arg_type}")


def construct_dataclass(cls: Dataclass_t, data: dict) -> 'DataclassInstance':
    """
    Construct (potentially recursively) a dataclass instance from a dictionary.

    Args:
        dataclass_type (type[DataclassInstance]): The dataclass type to construct
        data (dict): The dictionary to construct the dataclass instance from

    Returns:
        DataclassInstance: The constructed dataclass instance
    """
    fieldtypes = {f.name: f.type for f in cls.cls.__dataclass_fields__.values()}
    body = {}
    for field_name, raw_field_type in fieldtypes.items():
        if field_name in data:
            field_type = normalize_type(raw_field_type)
            if is_structured_type(field_type):
                body[field_name] = construct_structured_type(field_type, data[field_name])
            else:
                body[field_name] = data[field_name]
    return cls.cls(**body)


def construct_union(union: Union_t, data: dict) -> Any:
    """
    Construct a union type from a dictionary.

    Args:
        union (Union_t): The union type to construct
        data (dict): The dictionary to construct the union type from

    Returns:
        Any: The constructed union type
    """
    matches = []
    for t in union.types:
        if is_structured_type(t) and isinstance(data, dict):
            try:
                matches.append(construct_structured_type(t, data))
            except:
                continue
        else:
            try:
                matches.append(t.new(data))
            except:
                continue
    if len(matches) == 0:
        raise ValueError(f"Input data does not match any of the types in the union {union}")
    if len(matches) > 1:
        raise ValueError(f"Input data matches multiple types in the union {union}. Matches: {matches}")
    return matches[0]


def get_structured_input_description(arg_type: NormalizedType, arg_name: str, arg_desc: str, raw_arg_default: Any | None, *, indent: int) -> list[str]:
    """
    Generate the tool description for a structured argument like a dataclass or pydantic model.

    Args:
        arg_type (NormalizedType): The type of the structured argument. currently only supports dataclasses and pydantic models
        arg_name (str): The name of the argument
        arg_desc (str): The description of the argument
        raw_arg_default (Any|None): The default value of the argument. None indicates no default value
    """
    # convert the default value to a string
    if is_dataclass(raw_arg_default) and not isinstance(raw_arg_default, type):
        arg_default = str(asdict(raw_arg_default))  # convert dataclass to dict
    elif isinstance(raw_arg_default, BaseModel):
        arg_default = str(raw_arg_default.model_dump())  # convert pydantic model to dict
    else:
        arg_default = str(raw_arg_default) if raw_arg_default is not None else ''

    if isinstance(arg_type, Dataclass_t):
        return get_dataclass_input_description(arg_type, arg_name, arg_desc, arg_default, indent=indent)

    if isinstance(arg_type, PydanticModel_t):
        return get_pydantic_input_description(arg_type, arg_name, arg_desc, arg_default, indent=indent)

    if isinstance(arg_type, Tuple_t):
        return get_tuple_structured_input_description(arg_type, arg_name, arg_desc, arg_default, indent=indent)

    if isinstance(arg_type, (List_t, Union_t, Dict_t)):
        return get_composite_structured_input_description(arg_type, arg_name, arg_desc, arg_default, indent=indent)

    raise ValueError(f"Unsupported structured type {arg_name}: {arg_type}")


def get_dataclass_input_description(arg_type: Dataclass_t, arg_name: str, arg_desc: str, arg_default: str, *, indent: int) -> list[str]:
    """
    Build the input description for a dataclass, including all of its fields (and potentially their nested fields).

    Args:
        arg_type (Dataclass_t): The dataclass type to get the input description for
        arg_name (str): The name of the argument
        arg_desc (str): The description of the argument
        arg_default (str): The default value of the argument
        indent (int): The level of indentation to use for the description

    Returns:
        list[str]: A list of strings that make up the input description for the dataclass
    """
    chunks = []
    num_fields = len(arg_type.cls.__dataclass_fields__)
    num_required_fields = sum(1 for field in arg_type.cls.__dataclass_fields__.values() if isinstance(
        field.default, _MISSING_TYPE) and isinstance(field.default, _MISSING_TYPE))
    num_optional_fields = num_fields - num_required_fields

    # argument name and high level description
    chunks.append(f'\n{TAB*indent}"{arg_name}":')
    if arg_desc:
        chunks.append(f" {arg_desc}.")
    if arg_default:
        chunks.append(f" Defaults to {arg_default}.")
    if num_required_fields == 0:
        chunks.append(" a json object with zero or more of the following optional fields:\n")
    elif num_optional_fields > 0:
        chunks.append(" a json object with the following fields (optional fields may be omitted):\n")
    else:
        chunks.append(" a json object with the following fields:\n")

    # opening brackets
    chunks.append(f"{TAB*(indent)}{{")

    # get the description for each field
    for field_name, field in arg_type.cls.__dataclass_fields__.items():
        field_type = normalize_type(field.type)
        field_desc = field.metadata.get("description", "")

        # determine the default value of the field
        if not isinstance(field.default, _MISSING_TYPE):
            field_default = field.default
        elif not isinstance(field.default_factory, _MISSING_TYPE):
            field_default = field.default_factory()
        else:
            field_default = None

        # if the field is a structured type, recursively get the input description
        if is_structured_type(field_type):
            chunks.extend(get_structured_input_description(
                field_type, field_name, field_desc, field_default, indent=indent+1))
            continue

        # add the field description to the chunks
        chunks.append(
            f'\n{TAB*(indent+1)}"{field_name}": ({field_type}{", optional" if field_default is not None else ""})')
        if field_desc:
            chunks.append(f" {field_desc}.")
        if field_default is not None:
            chunks.append(f" Defaults to {field_default}.")

    # closing brackets
    chunks.append(f"\n{TAB*indent}}}")

    return chunks


def get_pydantic_input_description(arg_type: PydanticModel_t, arg_name: str, arg_desc: str, arg_default: str, *, indent: int) -> list[str]:
    """
    Build the input description for a pydantic model, including all of its fields (and potentially their nested fields).

    Args:
        arg_type (PydanticModel_t): The pydantic model type to get the input description for
        arg_name (str): The name of the argument
        arg_desc (str): The description of the argument
        arg_default (str): The default value of the argument
        indent (int): The level of indentation to use for the description

    Returns:
        list[str]: A list of strings that make up the input description for the pydantic model
    """
    chunks = []
    num_fields = len(arg_type.cls.model_fields)
    num_required_fields = sum(1 for field in arg_type.cls.model_fields.values() if field.is_required())
    num_optional_fields = num_fields - num_required_fields

    # argument name and high level description
    chunks.append(f'\n{TAB*indent}"{arg_name}":')
    if arg_desc:
        chunks.append(f" {arg_desc}.")
    if arg_default:
        chunks.append(f" Defaults to {arg_default}.")
    if num_required_fields == 0:
        chunks.append(" a json object with zero or more of the following optional fields:\n")
    elif num_optional_fields > 0:
        chunks.append(" a json object with the following fields (optional fields may be omitted):\n")
    else:
        chunks.append(" a json object with the following fields:\n")

    # opening brackets
    chunks.append(f"{TAB*(indent)}{{")

    # get the description for each field
    for field_name, field in arg_type.cls.model_fields.items():
        field_type = normalize_type(field.annotation)
        field_desc = field.description or ''

        # determine the default value of the field
        field_default = None
        if not field.is_required():
            if field.default_factory is not None:
                field_default = field.default_factory()
            else:
                field_default = field.default

        # if the field is a structured type, recursively get the input description
        if is_structured_type(field_type):
            chunks.extend(get_structured_input_description(
                field_type, field_name, field_desc, field_default, indent=indent+1))
            continue

        # add the field description to the chunks
        chunks.append(
            f'\n{TAB*(indent+1)}"{field_name}": ({field_type}{", optional" if field_default is not None else ""})')
        if field_desc:
            chunks.append(f" {field_desc}.")
        if field_default is not None:
            chunks.append(f" Defaults to {field_default}.")

    # closing brackets
    chunks.append(f"\n{TAB*indent}}}")

    return chunks



def get_composite_structured_input_description(arg_type: List_t | Dict_t | Union_t, arg_name: str, arg_desc: str, arg_default: str, *, indent: int) -> list[str]:
    """
    Build the input description for a composite structured type (List, Dict, or Union), including all of its fields (and potentially their nested fields).

    Args:
        arg_type (List_t|Dict_t|Union_t): The composite structured type to get the input description for
        arg_name (str): The name of the argument
        arg_desc (str): The description of the argument
        arg_default (str): The default value of the argument
        indent (int): The level of indentation to use for the description

    Returns:
        list[str]: A list of strings that make up the input description for the composite structured type
    """
    chunks = []
    structs = get_all_nested_structured_types(arg_type)

    chunks.append(f'\n{TAB*indent}"{arg_name}":')
    if arg_desc:
        chunks.append(f" {arg_desc}.")
    if arg_default:
        chunks.append(f" Defaults to {arg_default}.")
    chunks.append(f" Input should match type `{arg_type}` where:")
    for struct in structs:
        if isinstance(struct, Dataclass_t):
            chunks.extend(get_dataclass_input_description(struct, f'{struct}', '', '', indent=indent+1))
        elif isinstance(struct, PydanticModel_t):
            chunks.extend(get_pydantic_input_description(struct, f'{struct}', '', '', indent=indent+1))
        else:
            raise ValueError(f"INTERNAL ERROR: get_all_nested_structured_types is only supposed to return Dataclass_t and PydanticModel_t instances. Found '{struct}'")

    return chunks


def get_tuple_structured_input_description(arg_type: Tuple_t, arg_name: str, arg_desc: str, arg_default: str, *, indent: int) -> list[str]:
    raise NotImplementedError("Tuple structured types are not yet supported")



def get_all_nested_structured_types(arg_type: NormalizedType) -> set[Dataclass_t|PydanticModel_t]:
    """
    Collect all Dataclass_t and PydanticModel_t types that may be nested within a composite structured type.
    
    Args:
        arg_type (NormalizedType): The composite structured type to collect nested types from

    Returns:
        set[Dataclass_t|PydanticModel_t]: A set of all nested Dataclass_t and PydanticModel_t types
    """
    s: set[Dataclass_t|PydanticModel_t] = set()

    # Dataclass_t and PydanticModel_t are what we are looking for
    if isinstance(arg_type, Dataclass_t) or isinstance(arg_type, PydanticModel_t):
        s.add(arg_type)

    # Tuple_t look at any of the component types that are not ellipsis
    if isinstance(arg_type, Tuple_t):
        if not isinstance(arg_type.component_types, NotProvided):
            for t in arg_type.component_types: 
                if t != ...:
                    s.update(get_all_nested_structured_types(t)) 

    # List_t look at the element type
    if isinstance(arg_type, List_t):
        if not isinstance(arg_type.element_type, NotProvided):
            s.update(get_all_nested_structured_types(arg_type.element_type))

    # Dict_t look at the key and value types
    if isinstance(arg_type, Dict_t):
        if not isinstance(arg_type.key_type, NotProvided):
            s.update(get_all_nested_structured_types(arg_type.key_type))
        if not isinstance(arg_type.value_type, NotProvided):
            s.update(get_all_nested_structured_types(arg_type.value_type))

    # Union_t look at all the types
    if isinstance(arg_type, Union_t):
        if not isinstance(arg_type.types, NotProvided):
            for t in arg_type.types:
                s.update(get_all_nested_structured_types(t))

    return s


import logging
import re
from .constants import TAB
from .agent import Agent
from .archytypes import evaluate_type_str, normalize_type, NormalizedType, is_primitive_type, is_structured_type
from .structured_data_utils import get_structured_input_description, construct_structured_type
from typing import Callable, Any, ParamSpec, TypeVar, overload
from textwrap import indent
import inspect
from docstring_parser import parse as parse_docstring

from textwrap import indent
from types import FunctionType
from typing import Callable, Any


logger = logging.getLogger(__name__)


# TODO: separate system tools from library tools from user tools
#      ideally system tools will have no library dependencies

# TODO: parse extra long manual page from doc string. man_page can be seen by calling man <tool_name>
# man_page:str|None=None
# wrapper._man_page = man_page


# Class/type definition for types used in dependency injection.
AgentRef = type("AgentRef", (), {})
ToolNameRef = type("ToolNameRef", (), {})
ToolFnRef = type("ToolFnRef", (), {})
LoopControllerRef = type("LoopControllerRef", (), {})
ReactContextRef = type("ReactContextRef", (), {})

INJECTION_MAPPING = {
    AgentRef: "agent",
    ToolNameRef: "tool_name",
    ToolFnRef: "raw_tool",
    LoopControllerRef: "loop_controller",
    ReactContextRef: "react_context",
}


def toolset(*args, **kwargs):
    """
    A dummy decorator for backwards compatibility.
    Provides no funcitonality.
    Any class can now contain tools without a decorator.
    """
    logger.warning(
        "Warning: The usage of the @toolset decorator is deprecated and the decorator will be removed in a future version.")

    def decorator(cls):
        return cls
    return decorator


R = TypeVar("R")
P = ParamSpec("P")
@overload
def tool(func: Callable[P, R], /) -> Callable[P, R]: ...


@overload
def tool(*, name: str | None = None, autosummarize: bool = False,
         devmode: bool = False) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


@overload
def tool(func: Callable[P, R], /, *, name: str | None = None,
         autosummarize: bool = False, devmode: bool = False) -> Callable[P, R]: ...


def tool(func=None, /, *, name: str | None = None, autosummarize: bool = False, devmode: bool = False):
    """
    Decorator to convert a function into a tool for ReAct agents to use.

    Usage:
    ```
        @tool
        def my_tool(arg:type) -> type:
            '''
            Short description of the tool

            Long description of the tool

            Args:
                arg (type): Description of the argument

            Returns:
                type: Description of the return value

            Examples:
                optional description of the example
                >>> my_tool(arg)
                result
            '''
    ```
    """

    # decorator case where the decorator is used directly on the func
    # either `@tool def func()` or `tool(func, name='name', autosummarize=True)`
    if func is not None:
        return tool(name=name, autosummarize=autosummarize, devmode=devmode)(func)

    def decorator(func: Callable):
        # check that the decorator is being applied to a function
        if not inspect.isfunction(func):
            raise TypeError(
                f"tool decorator can only be applied to functions or classes. Got {func} of type {type(func)}"
            )

        func._name = name if name else func.__name__    # type: ignore
        func._is_tool = True                            # type: ignore
        func.autosummarize = autosummarize              # type: ignore
        func._devmode = devmode                         # type: ignore

        # attach usage description to the wrapper function
        args_list, ret, desc, injections = get_tool_signature(func)
        arg_preprocessor = make_arg_preprocessor(args_list)

        async def run(
            args: dict | None,
            tool_context: dict[str, object] = {},
            self_ref: object = None

        ):
            """Output from LLM will be dumped into a json object. Depending on object type, call func accordingly."""

            # Initialise positional and keyword argument structs
            pargs, kwargs = arg_preprocessor(args)

            if self_ref:
                pargs.insert(0, self_ref)

            # Add injections to kwargs
            for inj_name, inj_type in injections.items():
                context_key = INJECTION_MAPPING.get(inj_type, None)
                context_value = tool_context.get(context_key or '', None)
                if context_value:
                    kwargs[inj_name] = context_value

            if inspect.iscoroutinefunction(func):
                result = await func(*pargs, **kwargs)
            else:
                result = func(*pargs, **kwargs)

            # convert the result to a string if it is not already a string
            if not isinstance(result, str):
                result = str(result)

            return result

        # Add func as the attribute of the run method
        func.run = run  # type: ignore

        return func

    return decorator


def is_tool(obj: Callable | type) -> bool:
    """checks if an object is a tool function, tool method, tool class, or an instance of a class tool"""
    return (
        getattr(obj, '_is_tool', False)
    )


def get_tool_prompt_description(obj: Callable | type | Any):

    # get the list of arguments
    chunks = []
    if getattr(obj, '_disabled', False):
        return ""

    if inspect.isfunction(obj) or inspect.ismethod(obj):
        args_list, ret, desc, injections = get_tool_signature(obj)
        args_list = args_list
        ret_name, ret_type, ret_description = ret
        short_desc, long_desc, examples = desc

        ############### NAME/DESCRIPTION ###############
        chunks.append(f"{obj._name}:\n")    # type: ignore
        if short_desc:
            chunks.append(f"{TAB}{short_desc}\n\n")
        if long_desc:
            chunks.append(f"{indent(long_desc, TAB)}\n\n")

        #################### INPUT ####################
        chunks.append(f"{TAB}_input_: ")

        if len(args_list) == 0:
            chunks.append("None")

        # all other cases have arguments wrapped in a json
        else:
            chunks.append(f"a json object with the following fields:\n{TAB}{{")
            for arg_name, arg_type, arg_desc, arg_default in args_list:
                if is_structured_type(arg_type):
                    chunks.extend(get_structured_input_description(arg_type, arg_name, arg_desc or '', arg_default, indent=2))
                    continue

                if not is_primitive_type(arg_type):
                    raise ValueError(f"Unsupported argument type {arg_type}")

                chunks.append(
                    f'\n{TAB}{TAB}"{arg_name}": ({arg_type}{", optional" if arg_default else ""}) {arg_desc}')
            chunks.append(f"\n{TAB}}}")

        #################### OUTPUT ####################
        chunks.append("\n    _output_: ")
        chunks.append(f"({ret_type}) {ret_description or ''}")

        ############### EXAMPLES ###############
        # TODO: examples need to be parsed...
    else:
        tool_methods = [
            tool_method
            for tool_method in collect_tools_from_object(obj)
            if not getattr(tool_method, "_disabled", False)
        ]
        if tool_methods:
            ############### NAME/DESCRIPTION ###############
            if inspect.isclass(obj):
                chunks.append(f"{obj.__name__} (class):\n")
            else:
                chunks.append(f"{obj.__class__.__name__} (class instance):\n")
            docstring = inspect.getdoc(obj)
            if docstring:
                chunks.append(f"{indent(docstring, TAB)}\n\n")

            #################### METHODS ####################
            chunks.append(f"{TAB}methods:\n")
            for method in tool_methods:
                method_str = get_tool_prompt_description(method)
                chunks.append(f"{indent(method_str, TAB*2)}\n\n")

    return "".join(chunks)


def get_tool_signature(
    func: Callable,
) -> tuple[
    list[tuple[str, NormalizedType, str | None, str | None]],
    tuple[str | None, NormalizedType, str | None],
    tuple[str | None, str | None, list[str]],
    dict[str, type],
]:
    """
    Check that the docstring and function signature match for a tool function, and return all function information.

    Args:
        func (function): The function to check and extract information from

    Returns:
        args_list: A list of tuples (name, type, description, default) for each argument
        ret: A tuple (name, type, description) for the return value
        desc: A tuple (short_description, long_description, examples) from the docstring for the function
        injected_args: A dictionary of injected arguments and their types
    """
    assert inspect.isfunction(func) or inspect.ismethod(
        func), f"get_tool_signature can only be used on functions or methods. Got {func}"

    # get the function signature from the function and the docstring
    assert func.__doc__ is not None, f"Function '{func.__name__}' has no docstring. All tools must have a docstring that matches the function signature."
    docstring = parse_docstring(func.__doc__)
    signature = inspect.signature(func)

    # Extract argument information from the docstring
    docstring_args: dict[str, tuple[NormalizedType, str | None, str | None]] = {
        arg.arg_name: (
            evaluate_type_str(arg.type_name or '', func.__globals__, devmode=func._devmode),  # type: ignore
            arg.description,
            arg.default
        )
        for arg in docstring.params
    }

    # Extract argument information from the signature (ignore self from class methods)
    all_args = {
        k: v.annotation
        for i, (k, v) in enumerate(signature.parameters.items())
        if not (i == 0 and k == "self")
    }

    # Extract argument information from the signature (ignore self from class methods)
    injected_args = {k: v for k, v in all_args.items() if v in INJECTION_MAPPING}
    signature_args = {k: normalize_type(v) for k, v in all_args.items() if k not in injected_args}

    # Check if the docstring argument names match the signature argument names
    if set(docstring_args.keys()) != set(signature_args.keys()):
        raise ValueError(
            f"Docstring argument names do not match function signature argument names for function '{func.__name__}'. "
            f"Docstring args: {[*docstring_args.keys()]}, Signature args: {[*signature_args.keys()]}"
        )

    # Check if the docstring argument types match the signature argument types
    for arg_name, arg_type in signature_args.items():
        docstring_arg_type, _, _ = docstring_args[arg_name]
        if arg_type != docstring_arg_type:
            logger.warning((
                f"Docstring type '{docstring_arg_type}' does not match function signature type '{arg_type}' "
                f"for argument '{arg_name}' for function '{func.__name__}'"
            ))

    # Generate a list of tuples (name, type, description, default) for each argument
    # TODO: use the signature to determine if an argument is optional or not
    args_list = [
        (arg_name, signature_args[arg_name], arg_desc, arg_default)
        for arg_name, (arg_type, arg_desc, arg_default) in docstring_args.items()
    ]

    # get the return type and description/name if they exist
    ret_type = normalize_type(signature.return_annotation if signature.return_annotation != inspect._empty else None)
    if docstring.returns is None:
        docstring_ret_type = normalize_type(None)
        docstring_return_name = None
        docstring_return_description = None
    else:
        docstring_ret_type = evaluate_type_str(docstring.returns.type_name or '',
                                               func.__globals__, devmode=func._devmode)  # type: ignore
        docstring_return_name = docstring.returns.return_name
        docstring_return_description = docstring.returns.description

    ret = (docstring_return_name, ret_type, docstring_return_description)

    if ret_type != docstring_ret_type:
        logger.warning(
            f"Docstring return type '{docstring_ret_type}' does not match function signature return type '{ret_type}' for function '{func.__name__}'"
        )

    # get the docstring description and examples
    examples = [example.description for example in docstring.examples if example.description]
    desc = (docstring.short_description, docstring.long_description, examples)

    return args_list, ret, desc, injected_args


def make_arg_preprocessor(args_list: list[tuple[str, NormalizedType, str | None, str | None]]) -> Callable[[dict | None], tuple[list, dict]]:
    """
    Make a preprocessor function that converts the agent's input into a tool into *pargs, **kwargs for the tool function

    Args:
        args_list: A list of tuples (name, type, description, default) for each argument

    Returns:
        preprocessor (args: Any) -> (pargs, kwargs):
    """

    def preprocessor(args: dict[str, Any] | None) -> tuple[list, dict]:
        """
        Argument preprocessor function for a tool function.

        Args:
            args (dict|None): The input arguments for the tool function. None if no arguments are provided.

        Returns:
            tuple[list, dict]: The positional arguments and keyword arguments for the tool. i.e. call `func(*pargs, **kwargs)`
        """
        if not isinstance(args, dict) and args is not None:
            raise TypeError(f"_input_ must be a dictionary or None. Got {type(args)}")

        # zero argument case
        if len(args_list) == 0:
            assert args is None, f"Expected no arguments, got {args}"
            return [], {}
        assert args is not None, f"Expected arguments, got None"

        # general case, arguments wrapped in json. need to determine which need to be deserialized into structured types
        # TODO: this doesn't respect if a function signature has position-only vs keyword-only arguments. need to update get_tool_signature to include that information
        pargs = []
        kwargs = {}
        for arg_name, arg_type, _, _ in args_list:

            # potentially a default arg that is already provided.
            # if not, calling the tool will fail
            if arg_name not in args:
                continue

            # primitive types can be passed directly
            if is_primitive_type(arg_type):
                kwargs[arg_name] = args[arg_name]
                continue

            # structured types are deserialized from a dict into the dataclass/pydantic model
            if is_structured_type(arg_type):
                kwargs[arg_name] = construct_structured_type(arg_type, args[arg_name])
                continue

            raise ValueError(f"Unsupported structured type {arg_type}")

        return pargs, kwargs

    return preprocessor



def collect_tools_from_object(obj: object):
    result = []
    for item_name, item in inspect.getmembers(obj, predicate=lambda member: inspect.ismethod(member) or inspect.isfunction(member)):
        if is_tool(item):
            result.append(item)
    return result


def make_tool_dict(tools: list[Callable | type | Any]) -> dict[str, Callable]:
    """
    Create a dictionary of tools from a list of tool functions.

    Tries to use the '_name' attribute of the @tool function, otherwise uses the function name.

    Args:
        tools (list[Callable|type|Any]): A list of tool functions, or tool classes, or instances of tool classes.

    Returns:
        dict[str, Callable]: A dictionary of tools. Class methods of class tools are included as separate functions.
    """
    tool_dict = {}
    for tool in tools:
        # If the tool is actually a class and not an instance or function, instantiate it
        if isinstance(tool, type):
            tool = tool()
        if is_tool(tool):
            name = getattr(tool, "_name", None)
            if name is None and isinstance(tool, Agent):
                name = tool.__class__.__name__
            tool_dict[name] = tool

        # add each method to the tool dictionary under the name 'class_name.method_name'
        methods = inspect.getmembers(
            tool,
            predicate=lambda member: inspect.ismethod(member) or inspect.isfunction(member)
        )
        for _, method in methods:
            if is_tool(method):
                if isinstance(tool, type):
                    cls_name = getattr(tool, "_name", None) or tool.__name__
                else:
                    cls_name = getattr(tool, "_name", None) or tool.__class__.__name__
                method_name = getattr(method, "_name", None) or getattr(method, "__name__")
                method_name = f"{cls_name}.{method_name}"
                if method_name in tool_dict and method is not tool_dict[method_name]:
                    raise ValueError(f"Tool name '{method_name}' is already in use")
                tool_dict[method_name] = method
    return tool_dict


def get_tool_names(tools: list[Callable | type | Any]):
    res = list(make_tool_dict(tools).keys())
    return res


def test():
    from archytas.tools import ask_user, datetime_tool, timestamp
    from archytas.demo_tools import fib_n, example_tool, calculator, Jackpot

    for t in [
        ask_user,
        datetime_tool,
        timestamp,
        fib_n,
        example_tool,
        calculator,
        Jackpot,
    ]:
        print(get_tool_prompt_description(t))
        print()


def sanitize_toolname(name: str) -> str:
    # Function/tool names in OpenAI/Anthropic/etc messages must match the pattern '^[a-zA-Z0-9_-]+$'
    name = re.sub(r'[^a-zA-Z0-9_-]', '_', name)
    return name



if __name__ == "__main__":
    test()



from types import UnionType, NoneType, EllipsisType
from typing import Any, Optional, Union, Literal, List, Dict, Tuple, Iterable, get_origin, get_args, overload, TYPE_CHECKING
if TYPE_CHECKING:
    from _typeshed import DataclassInstance  # only available to type checkers
from abc import ABC, abstractmethod
from dataclasses import dataclass, is_dataclass
from pydantic import BaseModel
import logging

logger = logging.getLogger(__name__)





class NotProvided:
    """sentinel for optional parameters"""
    _instance = None
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    def __repr__(self):
        return 'NotProvided'
    def __str__(self):
        return 'NotProvided'
    def __eq__(self, other):
        return isinstance(other, NotProvided)
    def __hash__(self):
        return hash(NotProvided)

# instance of the singleton
notprovided = NotProvided()


@dataclass(frozen=True)
class NormalizedType(ABC):
    @abstractmethod
    def __str__(self) -> str: ...
    @classmethod
    def new(cls, value: Any) -> Any:
        """Primitive types can use .new() to create an instance of the type"""
        raise TypeError(f"Type Error: Cannot construct Type `{cls.__name__}.new()`")

# Too much of a hassle to make this one a dataclass since we need to flatten nested Union_t types
class Union_t(NormalizedType):
    # Union_t can take either an Iterable[NormalizedType] or multiple NormalizedType as arguments
    @overload
    def __init__(self, types: Iterable[NormalizedType]): ...
    @overload
    def __init__(self, *types: NormalizedType): ...
    def __init__(self, types: Iterable[NormalizedType]|NormalizedType, *args: NormalizedType):

        # convert args to a single set of NormalizedType
        if isinstance(types, NormalizedType):
            _types = {types, *args}
        else:
            assert len(args) == 0, "Union_t can only take one argument if it is an Iterable[NormalizedType]"
            _types = set(types)

        # flatten any nested Union_t types into a single layer
        while any(isinstance(t, Union_t) for t in _types):
            new_types = set()
            for t in _types:
                if isinstance(t, Union_t):
                    new_types.update(t.types)
                else:
                    new_types.add(t)
            _types = new_types
        
        self.types = frozenset(_types)

    def __str__(self) -> str:
        return ' | '.join(str(t) for t in self.types)
    
    def __repr__(self) -> str:
        return f"Union_t({set(self.types)})" # wrap with set() to print it nicer
    
    def __eq__(self, other):
        if not isinstance(other, Union_t):
            return NotImplemented
        return isinstance(other, Union_t) and self.types == other.types

    def __hash__(self):
        return hash(self.types)

@dataclass(frozen=True)
class List_t(NormalizedType):
    element_type: NormalizedType | NotProvided = notprovided
    
    def __str__(self) -> str:
        if isinstance(self.element_type, NotProvided):
            return 'list'
        return f'list[{self.element_type}]'


@dataclass(frozen=True)
class Tuple_t(NormalizedType):
    component_types: tuple[NormalizedType, ...] | tuple[NormalizedType, EllipsisType] | NotProvided = notprovided

    def __str__(self) -> str:
        if isinstance(self.component_types, NotProvided):
            return 'tuple'
        return f'tuple[{", ".join(str(t) for t in self.component_types)}]'


@dataclass(frozen=True)
class Dict_t(NormalizedType):
    #TODO: tbd if this is the best way to store key_type and value_type
    #      e.g. could be params: tuple[NormalizedType, NormalizedType] | NotProvided
    key_type: NormalizedType | NotProvided = notprovided
    value_type: NormalizedType | NotProvided = notprovided

    def __str__(self) -> str:
        if isinstance(self.key_type, NotProvided) and isinstance(self.value_type, NotProvided):
            return 'dict'
        if isinstance(self.key_type, NotProvided):
            return f'dict[?, {self.value_type}]'
        if isinstance(self.value_type, NotProvided):
            return f'dict[{self.key_type}, ?]'
        return f'dict[{self.key_type}, {self.value_type}]'


@dataclass(frozen=True)
class Int_t(NormalizedType):
    def __str__(self) -> str:
        return 'int'
    @classmethod
    def new(cls, value: Any) -> int:
        if not isinstance(value, (int, float)) or value != int(value):
            raise TypeError(f"Expected an int, got {value}")
        return int(value)


@dataclass(frozen=True)
class Float_t(NormalizedType):
    def __str__(self) -> str:
        return 'float'
    @classmethod
    def new(cls, value: Any) -> float:
        if not isinstance(value, (int, float)):
            raise TypeError(f"Expected a float, got {value}")
        return float(value)


@dataclass(frozen=True)
class Str_t(NormalizedType):
    def __str__(self) -> str:
        return 'str'
    @classmethod
    def new(cls, value: Any) -> str:
        if not isinstance(value, str):
            raise TypeError(f"Expected a string, got {value}")
        return value


@dataclass(frozen=True)
class Bool_t(NormalizedType):
    def __str__(self) -> str:
        return 'bool'
    @classmethod
    def new(cls, value: Any) -> bool:
        if not isinstance(value, bool):
            raise TypeError(f"Expected a bool, got {value}")
        return bool(value)


@dataclass(frozen=True)
class None_t(NormalizedType):
    def __str__(self) -> str:
        return 'None'
    @classmethod
    def new(cls, value: Any) -> None:
        if value is not None:
            raise TypeError(f"Expected None, got {value}")
        return None


@dataclass(frozen=True)
class Literal_t(NormalizedType): ... #TODO:


@dataclass(frozen=True)
class Dataclass_t(NormalizedType):
    cls: 'type[DataclassInstance]'
    def __post_init__(self):
        # TODO: check/normalize members?
        ...
    def __str__(self) -> str:
        return self.cls.__name__


@dataclass(frozen=True)
class PydanticModel_t(NormalizedType):
    cls: 'type[BaseModel]'
    def __post_init__(self):
        # TODO: check/normalize members?
        ...
    def __str__(self) -> str:
        return self.cls.__name__


# Any should compare equal to all types
@dataclass(frozen=True)
class Any_t(NormalizedType):
    def __str__(self) -> str:
        return 'Any'
    def __eq__(self, other):
        return True
    def __req__(self, other):
        return True


def is_optional(annotation):
    """Check if the annotation is typing.Optional[T]"""
    return get_origin(annotation) is Union and type(None) in get_args(annotation)


def normalize_type(t: Any) -> NormalizedType:
    """
    Convert a type annotation to a NormalizedType
    Supports old and new style type annotations
    e.g. `List[T]` vs `list[T]`, `Optional[T]` vs `T | None`, `Union[A, B]` vs `A | B`, etc.

    The following types are supported:
    - str, int, float, bool, None
    - list, dict, tuple
    - Optional, Union, (WIP) Literal
    - dataclasses
    - pydantic models

    Args:
        t (Any): the type annotation

    Returns:
        NormalizedType: the normalized type object
    """

    #### main error cases ####

    # Optional without a parameter doesn't make sense
    if t is Optional:
        raise ValueError("Underspecified type for tool. Optional should contain a type, e.g. Optional[int]. Note that Optional[T] is equivalent to Union[T, None]")
    
    # Object_t shouldn't be used! It's not really useful for agent instructions
    if t is object:
        # return Object_t()
        raise ValueError("Underspecified type annotation for tool. `object` does not provide enough information for the agent to create an instance of the argument to the tool.")

    # Any is also not useful for the same reasons as object
    if t is Any:
        raise ValueError("Underspecified type annotation for tool. `Any` does not provide enough information for the agent to create an instance of the argument to the tool.")


    # Primitive types
    if t is str:
        return Str_t()
    if t is int:
        return Int_t()
    if t is float:
        return Float_t()
    if t is bool:
        return Bool_t()
    if t is None or t is NoneType:
        return None_t()
    

    # Optional[a]
    if is_optional(t):
        return Union_t(normalize_type(a) for a in get_args(t))
    
    # a | b
    if isinstance(t, UnionType):
        return Union_t(normalize_type(a) for a in t.__args__)
    # Union[a, b]
    if get_origin(t) is Union:
        return Union_t(normalize_type(a) for a in get_args(t))

    # Literal[a, b, c]
    if get_origin(t) is Literal:
        raise NotImplementedError("Literal type annotation is not yet supported")
        # return Literal_t()
    
    #Tuple[a, b, c], Tuple, tuple
    if t is tuple or t is Tuple:
        return Tuple_t()
    if get_origin(t) is tuple:
        return Tuple_t(tuple(normalize_type(a) for a in get_args(t) if a != ...))
    
    #List[a], List, list
    if t is list or t is List:
        return List_t()
    if get_origin(t) is list:
        args = get_args(t)
        if len(args) != 1:
            raise ValueError(f"List type annotation should have exactly one argument, found {len(args)}")
        return List_t(normalize_type(args[0]))


    #Dict[a, b], Dict, dict
    if t is dict or t is Dict:
        return Dict_t()
    if get_origin(t) is dict:
        args = get_args(t)
        if len(args) != 2:
            raise ValueError(f"Dict type annotation should have exactly two arguments, found {len(args)}")
        return Dict_t(normalize_type(args[0]), normalize_type(args[1]))


    if isinstance(t, type) and is_dataclass(t): # is_dataclass also return True for dataclass instances, which we don't want to match
        #TODO: TBD if the inner types should be normalized...
        return Dataclass_t(t)

    if isinstance(t, type) and issubclass(t, BaseModel):
        #TODO: TBD if the inner types should be normalized...
        return PydanticModel_t(t)


    raise ValueError(f"Unsupported type to normalize: {t}")


#TODO: the only real way to make this safe is to manually parse the types ourselves
#      with `eval`, even if you restrict __builtins__ / etc., the user can still get access 
#      to them through the the function the docstring is attached to, e.g. `func.__globals__`
#      TBD on how necessary since generally we should be in control of the docstrings
#      and if we're not, the user would be able to execute arbitrary code anyways
# For now, the main mitigation is the devmode=False flag. Production code should always
#      have this set to False (forcing all docstring types to Any_t). When true, this
#      function will be called on the docstring's type annotations to normalize them.
def evaluate_type_str(type_str: str, globals: dict[str, Any], *, devmode: bool) -> NormalizedType:
    """
    Convert a string type annotation (i.e. from a docstring) into the actual normalized type object

    Args:
        type_str (str): the string type annotation
        globals (dict[str, Any]): the globals from the function that the annotation is attached to.
        devmode (bool): Only perform evaluation in devmode mode. If False (i.e. production), returns Any_t.

    Returns:
        NormalizedType: the normalized type object. If devmode is False, this will always be Any_t.
    """
    # Short circuit for non-devmode (e.g. production)
    if not devmode:
        return Any_t()

    # evaluate the type from the docstring in the context of the function that defined it
    try:
        t = eval(type_str, globals)
    except Exception as e:
        raise ValueError(f"Could not evaluate type string '{type_str}'") from e

    return normalize_type(t)



def is_primitive_type(arg_type: NormalizedType) -> bool:
    """
    Check if a type is a primitive type

    Primitive types are `str`, `int`, `float`, `bool`, `list`, and `dict`
    Additionally list and dict may be parameterized with primitive types. e.g. `list[str]`, `dict[str, int]`
    Lastly unions are considered primitive if all of their arguments are primitive types. e.g. `str | int`
    """

    # simplest case
    if isinstance(arg_type, (Str_t, Int_t, Float_t, Bool_t, None_t)):
        return True

    # for list, dict, tuple, and union: any inner types must be primitive
    if isinstance(arg_type, List_t):
        if isinstance(arg_type.element_type, NotProvided):
            return True
        return is_primitive_type(arg_type.element_type)

    if isinstance(arg_type, Tuple_t):
        if isinstance(arg_type.component_types, NotProvided):
            return True
        return all(is_primitive_type(t) for t in arg_type.component_types if t != ...)

    if isinstance(arg_type, Dict_t):
        if isinstance(arg_type.key_type, NotProvided) and isinstance(arg_type.value_type, NotProvided):
            return True
        if not isinstance(arg_type.key_type, NotProvided) and not is_primitive_type(arg_type.key_type):
            return False
        if not isinstance(arg_type.value_type, NotProvided) and not is_primitive_type(arg_type.value_type):
            return False
        return True

    if isinstance(arg_type, Union_t):
        return all(is_primitive_type(t) for t in arg_type.types)

    return False


def is_structured_type(arg_type: NormalizedType) -> bool:
    """Check if a type is a structured type like a dataclass or pydantic model"""
    if isinstance(arg_type, Dataclass_t) or isinstance(arg_type, PydanticModel_t):
        return True

    if isinstance(arg_type, Union_t):
        return any(is_structured_type(t) for t in arg_type.types)

    if isinstance(arg_type, List_t):
        if not isinstance(arg_type.element_type, NotProvided):
            return is_structured_type(arg_type.element_type)

    if isinstance(arg_type, Tuple_t):
        if not isinstance(arg_type.component_types, NotProvided):
            return any(is_structured_type(t) for t in arg_type.component_types if t != ...)

    if isinstance(arg_type, Dict_t):
        if not isinstance(arg_type.key_type, NotProvided):
            return is_structured_type(arg_type.key_type)
        if not isinstance(arg_type.value_type, NotProvided):
            return is_structured_type(arg_type.value_type)

    return False



import json
import re
import logging

from anthropic import AuthenticationError as AnthropicAuthenticError, RateLimitError
from langchain_anthropic.chat_models import ChatAnthropic
from langchain_core.messages import AIMessage, SystemMessage
from pydantic import BaseModel as PydanticModel, Field

from archytas.agent import AIMessage, BaseMessage

from .base import BaseArchytasModel, EnvironmentAuth, ModelConfig
from ..exceptions import AuthenticationError, ExecutionError
from ..utils import extract_json

class DummyTool(PydanticModel):
    """
    Dummy Tool
    """
    input: str = Field(..., description="input")


class AnthropicModel(BaseArchytasModel):
    api_key: str = ""

    def __init__(self, config: ModelConfig, **kwargs) -> None:
        super().__init__(config, **kwargs)
        self.last_messages: list[BaseMessage] = None

    def auth(self, **kwargs) -> None:
        if 'api_key' in kwargs:
            self.api_key = kwargs['api_key']
        elif 'api_key' in self.config:
            self.api_key = self.config['api_key']
        if not self.api_key:
            raise AuthenticationError("No auth credentials found.")

    def initialize_model(self, **kwargs):
        return ChatAnthropic(model=self.config.get("model_name", "claude-2.1"), api_key=self.api_key).bind_tools([
            DummyTool
        ])

    def _preprocess_messages(self, messages):
        from ..agent import AutoContextMessage, ContextMessage
        output = []
        system_messages = []
        for message in messages:
            match message:
                case SystemMessage() | ContextMessage() | AutoContextMessage():
                    system_messages.append(message.content)
                case AIMessage():
                    # Duplicate mesage so we don't change raw storage
                    msg = message.model_copy(deep=True)
                    if not isinstance(msg.content, list):
                        content = [
                            {
                                "type": "text",
                                "text": msg.content,
                            }
                        ]
                        msg.content = content
                    for tool_call in msg.tool_calls:
                        tool_name = re.sub(r'[^a-zA-Z0-9_-]', '_', tool_call.get("name", "DummyTool"))
                        if tool_name not in [m["name"] for m in msg.content if m["type"] == "tool_use"]:
                            msg.content.append({
                                "type": "tool_use",
                                "id": tool_call.get("id"),
                                "name": tool_name,
                                "input": tool_call.get("args"),
                            })
                    msg.tool_calls = []
                    output.append(msg)
                case _:
                    output.append(message)
        # Condense all context/system messages into a single first message as required by Anthropic
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        self.last_messages = [msg.model_copy(deep=True) for msg in output]
        return output

    def _rectify_result(self, response_message: AIMessage):
        if isinstance(response_message.content, list):
            response_message.content = "\n".join(item.get("text") for item in response_message.content if item.get("type") == "text")
            response_message.tool_calls = []
        return response_message

    def process_result(self, response_message: AIMessage):
        content = super().process_result(response_message)
        try:
            result = extract_json(content)
            if isinstance(result, list):
                # Only perform the first action
                return result[0]
        except Exception as e:
            result = content
        return result

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, AnthropicAuthenticError):
            raise AuthenticationError("Anthropic Authentication Error") from error
        # TODO: Retry with delay on rate limit errors?
        # elif isinstance(error, RateLimitError):
        #     raise
        else:
            message_output = [msg.model_dump() for msg in self.last_messages]
            logging.warning(
                "An exception has occurred. Below are the messages that were sent to in the most recent request:\n" +
                json.dumps(message_output, indent=2)
            )
            raise
